@@ -27,6 +27,15 @@ from typing import Iterable
 from typing import Any
 import sys
 import ctypes
+import logging
+from gi import require_version # type: ignore
+# pylint: disable=wrong-import-position
+require_version('IBus', '1.0')
+from gi.repository import IBus # type: ignore
+# pylint: enable=wrong-import-position
+import itb_util
+
+LOGGER = logging.getLogger('ibus-typing-booster')
 
 # pylint: disable=invalid-name
 # pylint: disable=too-few-public-methods
@@ -908,6 +917,7 @@ def convert_digits_to_ascii(text: str) -> str:
     return text.translate(DIGIT_TRANS_TABLE)
 
 class Transliterator:
+    # pylint: disable=line-too-long
     '''A class for transliterators using libm17n
 
     If initializing the transliterator fails, for example because a
@@ -941,31 +951,23 @@ class Transliterator:
     'नमस्ते'
 
     >>> trans.transliterate_parts(list('n'))
-    ('', 'न्')
-
+    ('', 0, 'न्', 2, 'क', [])
     >>> trans.transliterate_parts(list('n '))
-    ('न ', '')
-
+    ('न ', 2, '', 0, 'क', [])
     >>> trans.transliterate_parts(list('na'))
-    ('', 'न')
-
+    ('', 0, 'न', 1, 'क', [])
     >>> trans.transliterate_parts(list('nam'))
-    ('न', 'म्')
-
+    ('न', 2, 'म्', 2, 'क', [])
     >>> trans.transliterate_parts(list('nama'))
-    ('न', 'म')
-
+    ('न', 2, 'म', 1, 'क', [])
     >>> trans.transliterate_parts(list('namas'))
-    ('नम', 'स्')
-
+    ('नम', 4, 'स्', 2, 'क', [])
     >>> trans.transliterate_parts(list('namast'))
-    ('नम', 'स्त्')
-
+    ('नम', 4, 'स्त्', 4, 'क', [])
     >>> trans.transliterate_parts(list('namaste'))
-    ('नम', 'स्ते')
-
+    ('नम', 4, 'स्ते', 4, 'क', [])
     >>> trans.transliterate_parts(list('namaste '))
-    ('नमस्ते ', '')
+    ('नमस्ते ', 8, '', 0, 'क', [])
 
     >>> trans.transliterate(list('. '))
     '। '
@@ -1008,11 +1010,53 @@ class Transliterator:
     >>> trans.transliterate(list('annyeonghaseyo'))
     '안녕하세요'
 
-    >>> trans = Transliterator('si-wijesekera')
+    >>> trans = Transliterator('si-wijesekara')
     >>> trans.transliterate(list('vksIal kjSka '))
     'ඩනිෂ්ක නවීන් '
 
+    >>> trans = Transliterator('t-latn-post')
+    >>> trans.transliterate_parts(list('u'))
+    ('', 0, 'u', 1, 'Latin-post', [])
+    >>> trans.transliterate_parts(list('u"'))
+    ('', 0, 'ü', 1, 'Latin-post', [])
+    >>> trans.transliterate_parts(list('u""'))
+    ('u"', 3, '', 0, 'Latin-post', [])
+    >>> trans.transliterate_parts(list('u"u'))
+    ('ü', 2, 'u', 1, 'Latin-post', [])
+    >>> trans.transliterate_parts(list('üu"u'))
+    ('üü', 3, 'u', 1, 'Latin-post', [])
+
+    >>> trans = Transliterator('t-rfc1345')
+    >>> trans.transliterate_parts(list('&'))
+    ('', 0, '&', 1, 'RFC1345', [])
+    >>> trans.transliterate_parts(list('&C'))
+    ('', 0, '&C', 2, 'RFC1345', [])
+    >>> trans.transliterate_parts(list('&Co'))
+    ('©', 3, '', 0, 'RFC1345', [])
+    >>> trans.transliterate_parts(list('&f'))
+    ('', 0, '&f', 2, 'RFC1345', [])
+    >>> trans.transliterate_parts(list('&ff'))
+    ('', 0, 'ﬀ', 1, 'RFC1345', [])
+    >>> trans.transliterate_parts(list('&ffi'))
+    ('ﬃ', 4, '', 0, 'RFC1345', [])
+
+    >>> trans = Transliterator('t-lsymbol')
+    >>> trans.transliterate_parts(list('/:)'))
+    ('', 0, '☺️', 2, 'lsymbol', ['☺️', '😃', '😅', '😆', '😉', '😇', '😂', '😏', '😛', '😜', '😝', '😋', '😉', '💏', '💋', '😍', '😘', '😚', '😽', '😻'])
+    >>> trans.transliterate_parts(list('a'))
+    ('a', 1, '', 0, 'lsymbol', [])
+    >>> trans.transliterate_parts(list('a/'))
+    ('a', 1, '/', 1, 'lsymbol', [])
+    >>> trans.transliterate_parts(list('a/:'))
+    ('a', 1, '/:', 2, 'lsymbol', [])
+    >>> trans.transliterate_parts(list('a/:('))
+    ('a', 1, '😢', 1, 'lsymbol', ['😢', '😩', '😡', '😭', '😪', '🙈', '🙊', '🙉'])
+    >>> trans.transliterate_parts(list('a/:(b'))
+    ('a😢b', 5, '', 0, 'lsymbol', [])
+
+    For a test transliterating parts using 'ja-anthy' see 'tests/test_m17n_translit.py'.
     '''
+    # pylint: enable=line-too-long
     def __init__(self, ime: str) -> None:
         '''Initialize the input method to use for the transliteration
 
@@ -1045,7 +1089,10 @@ class Transliterator:
         except ValueError as error: # NULL pointer access
             raise ValueError('minput_create_ic() failed') from error
 
-    def transliterate_parts(self, msymbol_list: Iterable[str], ascii_digits: bool = False) -> Tuple[str, str]:
+    def transliterate_parts(
+            self,
+            msymbol_list: Iterable[str],
+            ascii_digits: bool = False) -> Tuple[str, int, str, int, str, List[str]]:
         '''Transliterate a list of Msymbol names
 
         :param msymbol_list: A list of strings which are interpreted
@@ -1055,16 +1102,48 @@ class Transliterator:
                              Msymbols is just joined to a single string.
         :param ascii_digits: If true, convert language specific digits
                              to ASCII digits
-        :return: The transliteration in two parts: (committed, preedit)
+        :return: The transliteration in several parts:
+
+        (committed, committed_index, preedit, cursor_pos, status, candidates)
+
+        committed: str        The part of the transliteration which cannot be
+                              changed anymore by adding more input, could be
+                              committed already if desired.
+        committed_index: int  The index up to which the msymbol_list input
+                              was “used up” to create the “committed” text.
+        preedit:              The transliteration of the remaining input,
+                              may still change by adding more input.
+        cursor_pos: int       The cursor position in the preedit.
+                              Counted in codepoints, not glyphs.
+                              Usually this is at the end of the preedit
+                              but an input method may move the cursor
+                              within the preedit!
+                              (I think only ja-anthy.mim actually uses this)
+        status: str           May change for some input methods to
+                              indicate a state.
+                              For example in case of ja-anthy.mim,
+                              this is 'aあ' before Henkan and changes
+                              to '漢' in Henkan mode.
+        candidates: List[str] May contain a list of candidates if the
+                              input method can produce multiple candidates.
         '''
         if not isinstance(msymbol_list, list):
             raise ValueError('Argument of transliterate() must be a list.')
         if self._dummy:
-            return (''.join(msymbol_list), '')
+            return (''.join(msymbol_list), 0, '', 0, '', [])
         libm17n__minput_reset_ic(self._ic) # type: ignore
         committed = ''
+        committed_index = 0
         preedit = ''
-        for symbol in msymbol_list:
+        candidates: List[str] = []
+        for index, symbol in enumerate(msymbol_list):
+            if len(symbol) == 1 and not itb_util.is_ascii(symbol):
+                symbol = IBus.keyval_name(IBus.unicode_to_keyval(symbol))
+            elif (len(symbol) == 3 and symbol[1] == '-'
+                and symbol[0] in ('G', 'C', 'A')
+                and not itb_util.is_ascii(symbol[2])):
+                symbol = symbol[:2] + IBus.keyval_name(
+                    IBus.unicode_to_keyval(symbol[2]))
             _symbol = libm17n__msymbol(symbol.encode('utf-8')) # type: ignore
             retval = libm17n__minput_filter( # type: ignore
                 self._ic, _symbol, ctypes.c_void_p(None))
@@ -1074,8 +1153,10 @@ class Transliterator:
                     self._ic, _symbol, ctypes.c_void_p(None), _mt)
                 if libm17n__mtext_len(_mt) > 0: # type: ignore
                     committed += mtext_to_string(_mt)
+                    committed_index = index
                 if retval:
-                    committed += symbol
+                    committed += msymbol_list[index]
+                    committed_index = index + 1
         try:
             if (self._ic.contents.preedit_changed
                 and
@@ -1085,6 +1166,42 @@ class Transliterator:
         except Exception as error: # pylint: disable=broad-except
             # This should never happen:
             raise ValueError('Problem accessing preedit') from error
+        plist = self._ic.contents.candidate_list
+        while bool(plist):  # NULL pointers have a False boolean value
+            key = libm17n__mplist_key(plist) # type: ignore
+            if not bool(key):
+                break
+            key_name = libm17n__msymbol_name(key.contents) # type: ignore
+            if key_name == b'mtext':
+                characters = mtext_to_string(
+                    ctypes.cast(libm17n__mplist_value(plist), # type: ignore
+                        ctypes.POINTER(libm17n__MText)))
+                candidates += list(characters)
+            elif key_name == b'plist':
+                candidate_plist = ctypes.cast(
+                    libm17n__mplist_value(plist), # type: ignore
+                    ctypes.POINTER(libm17n__MPlist))
+                while True:
+                    candidate_plist_key = libm17n__mplist_key( # type: ignore
+                        candidate_plist)
+                    if not bool(candidate_plist_key):
+                        break
+                    candidate_plist_key_name = libm17n__msymbol_name( # type: ignore
+                        candidate_plist_key.contents)
+                    if candidate_plist_key_name != b'mtext':
+                        break
+                    candidate = mtext_to_string(
+                        ctypes.cast(
+                            libm17n__mplist_value(candidate_plist), # type: ignore
+                            ctypes.POINTER(libm17n__MText)))
+                    candidates.append(candidate)
+                    candidate_plist = libm17n__mplist_next( # type: ignore
+                        candidate_plist)
+            else:
+                break
+            plist = libm17n__mplist_next(plist) # type: ignore
+        cursor_pos = self._ic.contents.cursor_pos
+        status = mtext_to_string(self._ic.contents.status)
         # From the m17n-lib documentation:
         #
         # The minput_reset_ic () function resets input context $IC by
@@ -1109,10 +1226,21 @@ class Transliterator:
         _symbol = libm17n__msymbol(b'nil') # type: ignore
         _retval = libm17n__minput_filter( # type: ignore
             self._ic, _symbol, ctypes.c_void_p(None))
+        if committed and not preedit:
+            committed_index = len(msymbol_list)
         if not ascii_digits:
-            return (committed, preedit)
+            return (committed,
+                    committed_index,
+                    preedit,
+                    cursor_pos,
+                    status,
+                    candidates)
         return (convert_digits_to_ascii(committed),
-                convert_digits_to_ascii(preedit))
+                committed_index,
+                convert_digits_to_ascii(preedit),
+                cursor_pos,
+                status,
+                candidates)
 
     def transliterate(self, msymbol_list: Iterable[str], ascii_digits: bool = False) -> str:
         '''Transliterate a list of Msymbol names
@@ -1126,7 +1254,13 @@ class Transliterator:
                              to ASCII digits
         :return: The transliteration in one string
         '''
-        (committed, preedit) = self.transliterate_parts(msymbol_list, ascii_digits)
+        (committed,
+         _committed_index,
+         preedit,
+         _cursor_pos,
+         _status,
+         _candidates) = self.transliterate_parts(
+            msymbol_list, ascii_digits)
         return committed + preedit
 
     def get_variables(self) -> List[Tuple[str, str, str]]:
@@ -1139,7 +1273,7 @@ class Transliterator:
         >>> trans = Transliterator('NoIME')
         >>> trans.get_variables()
         []
-        >>> trans = Transliterator('si-wijesekera')
+        >>> trans = Transliterator('si-wijesekara')
         >>> trans.get_variables()
         [('use-surrounding-text', 'Surrounding text vs. preedit.\\nIf 1, try to use surrounding text.  Otherwise, use preedit.', '0')]
         >>> trans = Transliterator('t-unicode')
@@ -1329,6 +1463,9 @@ class Transliterator:
         return
 
 if __name__ == "__main__":
+    LOG_HANDLER = logging.StreamHandler(stream=sys.stderr)
+    LOGGER.setLevel(logging.DEBUG)
+    LOGGER.addHandler(LOG_HANDLER)
     import doctest
     (FAILED, _ATTEMPTED) = doctest.testmod()
     sys.exit(FAILED)
